"""Multiplexer message handling."""
import attr

CHANNEL_FLOW_NEW = 0x01
CHANNEL_FLOW_DATA = 0x02
CHANNEL_FLOW_CLOSE = 0x04
CHANNEL_FLOW_PING = 0x08

CHANNEL_FLOW_ALL = [
    CHANNEL_FLOW_NEW, CHANNEL_FLOW_CLOSE, CHANNEL_FLOW_DATA, CHANNEL_FLOW_PING
]


@attr.s(frozen=True)
class MultiplexerMessage:
    """Represent a message from multiplexer stream."""

    channel_id = attr.ib(type=str)
    flow_type = attr.ib(
        type=int, validator=attr.validators.in_(CHANNEL_FLOW_ALL))
    data = attr.ib(type=bytes, default=b"")
    extra = attr.ib(type=bytes, default=b"")



"""Public peer interface."""
import asyncio
from contextlib import suppress
import logging
from typing import Optional

import async_timeout

from ..exceptions import SniTunChallengeError, SniTunInvalidPeer
from .peer_manager import PeerManager

_LOGGER = logging.getLogger(__name__)


class PeerListener:
    """Peer Listener class."""

    def __init__(self, peer_manager: PeerManager, host=None, port=None):
        """Initialize SNI Proxy interface."""
        self._peer_manager = peer_manager
        self._host = host
        self._port = port or 8080
        self._server = None

    async def start(self):
        """Start peer server."""
        self._server = await asyncio.start_server(
            self.handle_connection, host=self._host, port=self._port)

    async def stop(self):
        """Stop peer server."""
        self._server.close()
        await self._server.wait_closed()

    async def handle_connection(self,
                                reader: asyncio.StreamReader,
                                writer: asyncio.StreamWriter,
                                data: Optional[bytes] = None):
        """Internal handler for incoming requests."""
        if not data:
            try:
                async with async_timeout.timeout(2):
                    fernet_data = await reader.read(2048)
            except asyncio.TimeoutError:
                _LOGGER.warning("Close peer handshake because timeout")
                writer.close()
                return
        else:
            fernet_data = data

        peer = None
        try:
            # Connection closed before data received
            if not fernet_data:
                return

            peer = self._peer_manager.register_peer(fernet_data)

            # Start multiplexer
            await peer.init_multiplexer_challenge(reader, writer)
            await peer.wait_disconnect()

        except SniTunInvalidPeer:
            _LOGGER.debug("Close because invalid fernet data")

        except SniTunChallengeError:
            _LOGGER.debug("Close because challenge was wrong")

        finally:
            if peer:
                self._peer_manager.remove_peer(peer)
            with suppress(OSError):
                writer.close()



"""Multiplexer for SniTun."""
import asyncio
from contextlib import suppress
import ipaddress
import logging
import os
import uuid

import async_timeout

from ..exceptions import (MultiplexerTransportClose,
                          MultiplexerTransportDecrypt,
                          MultiplexerTransportError)
from ..utils.ipaddress import bytes_to_ip_address
from .channel import MultiplexerChannel
from .crypto import CryptoTransport
from .message import (CHANNEL_FLOW_CLOSE, CHANNEL_FLOW_DATA, CHANNEL_FLOW_NEW,
                      CHANNEL_FLOW_PING, MultiplexerMessage)

_LOGGER = logging.getLogger(__name__)


class Multiplexer:
    """Multiplexer Socket wrapper."""

    def __init__(self,
                 crypto: CryptoTransport,
                 reader: asyncio.StreamReader,
                 writer: asyncio.StreamWriter,
                 new_connections=None):
        """Initialize Multiplexer."""
        self._crypto = crypto
        self._reader = reader
        self._writer = writer
        self._loop = asyncio.get_event_loop()
        self._queue = asyncio.Queue(20)
        self._processing_task = self._loop.create_task(self._runner())
        self._channels = {}
        self._new_connections = new_connections

    @property
    def is_connected(self) -> bool:
        """Return True is they is connected."""
        return not self._processing_task.done()

    def wait(self) -> asyncio.Task:
        """Block until the connection is closed.

        Return a awaitable object.
        """
        return asyncio.shield(self._processing_task)

    async def shutdown(self):
        """Shutdown connection."""
        if self._processing_task.done():
            return

        _LOGGER.debug("Cancel connection")
        self._processing_task.cancel()

        await self._graceful_channel_shutdown()

    async def _graceful_channel_shutdown(self):
        """Graceful shutdown of channels."""
        tasks = [
            channel.message_transport(channel.init_close())
            for channel in self._channels.values()
        ]
        self._channels.clear()

        if tasks:
            await asyncio.wait(tasks)

    def ping(self):
        """Send a ping flow message to hold the connection open."""
        message = MultiplexerMessage(uuid.uuid4(), CHANNEL_FLOW_PING)
        self._queue.put_nowait(message)

    async def _runner(self):
        """Runner task of processing stream."""
        transport = self._writer.transport
        from_peer = None
        to_peer = None

        # Process stream
        try:
            while not transport.is_closing():
                if not from_peer:
                    from_peer = self._loop.create_task(
                        self._reader.readexactly(32))

                if not to_peer:
                    to_peer = self._loop.create_task(self._queue.get())

                # Wait until data need to be processed
                await asyncio.wait([from_peer, to_peer],
                                   return_when=asyncio.FIRST_COMPLETED)

                # From peer
                if from_peer.done():
                    if from_peer.exception():
                        raise from_peer.exception()
                    await self._read_message(from_peer.result())
                    from_peer = None

                # To peer
                if to_peer.done():
                    if to_peer.exception():
                        raise to_peer.exception()
                    self._write_message(to_peer.result())
                    to_peer = None

        except asyncio.CancelledError:
            _LOGGER.debug("Receive canceling")
            with suppress(OSError):
                self._writer.write_eof()
                await self._writer.drain()

        except (MultiplexerTransportClose, asyncio.IncompleteReadError):
            _LOGGER.debug("Transport was closed")

        finally:
            if to_peer and not to_peer.done():
                to_peer.cancel()
            if from_peer and not from_peer.done():
                from_peer.cancel()
            if not transport.is_closing():
                with suppress(OSError):
                    self._writer.close()

        await self._graceful_channel_shutdown()
        _LOGGER.debug("Multiplexer connection is closed")

    def _write_message(self, message: MultiplexerMessage) -> None:
        """Write message to peer."""
        header = message.channel_id.bytes
        header += message.flow_type.to_bytes(1, byteorder='big')
        header += len(message.data).to_bytes(4, byteorder='big')
        header += message.extra + os.urandom(11 - len(message.extra))

        data = self._crypto.encrypt(header) + message.data
        self._writer.write(data)

    async def _read_message(self, header: bytes) -> None:
        """Read message from peer."""
        if not header:
            raise MultiplexerTransportClose()

        try:
            header = self._crypto.decrypt(header)
            channel_id = header[:16]
            flow_type = header[16]
            data_size = int.from_bytes(header[17:21], byteorder='big')
            extra = header[21:]
        except (IndexError, MultiplexerTransportDecrypt):
            _LOGGER.warning("Wrong message header received")
            return

        # Read message data
        if data_size:
            data = await self._reader.readexactly(data_size)
        else:
            data = b""

        message = MultiplexerMessage(
            uuid.UUID(bytes=channel_id), flow_type, data, extra)

        # Process message to queue
        await self._process_message(message)

    async def _process_message(self, message: MultiplexerMessage) -> None:
        """Process received message."""

        # DATA
        if message.flow_type == CHANNEL_FLOW_DATA:
            # check if message exists
            if message.channel_id not in self._channels:
                _LOGGER.warning("Receive data from unknown channel")
                return
            await self._channels[message.channel_id].message_transport(message)

        # New
        elif message.flow_type == CHANNEL_FLOW_NEW:
            # Check if we would handle new connection
            if not self._new_connections:
                _LOGGER.warning("Request new Channel is not allow")
                return

            ip_address = bytes_to_ip_address(message.extra[1:5])
            channel = MultiplexerChannel(self._queue, ip_address,
                                         message.channel_id)
            self._channels[channel.uuid] = channel
            self._loop.create_task(self._new_connections(self, channel))

        # Close
        elif message.flow_type == CHANNEL_FLOW_CLOSE:
            # check if message exists
            if message.channel_id not in self._channels:
                _LOGGER.warning("Receive close from unknown channel")
                return
            channel = self._channels.pop(message.channel_id)
            await channel.message_transport(message)

        # Ping
        elif message.flow_type == CHANNEL_FLOW_PING:
            _LOGGER.debug("Receive ping from peer")

        else:
            _LOGGER.warning("Receive unknown message. Abort connection")
            self._loop.create_task(self.shutdown())

    async def create_channel(
            self, ip_address: ipaddress.IPv4Address) -> MultiplexerChannel:
        """Create a new channel for transport."""
        channel = MultiplexerChannel(self._queue, ip_address)
        message = channel.init_new()

        try:
            async with async_timeout.timeout(5):
                await self._queue.put(message)
        except asyncio.TimeoutError:
            raise MultiplexerTransportError() from None
        else:
            self._channels[channel.uuid] = channel

        return channel

    async def delete_channel(self, channel: MultiplexerChannel) -> None:
        """Delete channel from transport."""
        message = channel.init_close()

        try:
            async with async_timeout.timeout(5):
                await self._queue.put(message)
        except asyncio.TimeoutError:
            raise MultiplexerTransportError() from None
        finally:
            self._channels.pop(channel.uuid, None)



"""SniTun client for server connection."""
import asyncio
import hashlib
import logging

import async_timeout

from .connector import Connector
from ..multiplexer.crypto import CryptoTransport
from ..multiplexer.core import Multiplexer
from ..exceptions import MultiplexerTransportDecrypt, SniTunConnectionError

_LOGGER = logging.getLogger(__name__)


class ClientPeer:
    """Client to SniTun Server."""

    def __init__(self, snitun_host: str, snitun_port=None):
        """Initialize ClientPeer connector."""
        self._multiplexer = None
        self._loop = asyncio.get_event_loop()
        self._snitun_host = snitun_host
        self._snitun_port = snitun_port or 8080

    @property
    def is_connected(self) -> bool:
        """Return true, if a connection exists."""
        return self._multiplexer is not None

    def wait(self) -> asyncio.Task:
        """Block until connection to peer is closed."""
        if not self._multiplexer:
            raise RuntimeError("No SniTun connection available")
        return self._multiplexer.wait()

    async def start(self, connector: Connector, fernet_token: bytes,
                    aes_key: bytes, aes_iv: bytes) -> None:
        """Connect an start ClientPeer."""
        if self._multiplexer:
            raise RuntimeError("SniTun connection available")

        # Connect to SniTun server
        try:
            reader, writer = await asyncio.open_connection(
                host=self._snitun_host, port=self._snitun_port)
        except OSError:
            _LOGGER.error("Can't connect to SniTun server %s:%s",
                          self._snitun_host, self._snitun_port)
            raise SniTunConnectionError()

        # Send fernet token
        writer.write(fernet_token)
        await writer.drain()

        # Challenge/Response
        crypto = CryptoTransport(aes_key, aes_iv)
        try:
            challenge = await reader.readexactly(32)
            answer = hashlib.sha256(crypto.decrypt(challenge)).digest()

            writer.write(crypto.encrypt(answer))
            await writer.drain()
        except (MultiplexerTransportDecrypt, asyncio.IncompleteReadError,
                OSError):
            _LOGGER.error("Challenge/Response error with SniTun server")
            raise SniTunConnectionError()

        # Run multiplexer
        self._multiplexer = Multiplexer(crypto, reader, writer,
                                        connector.handler)

        # Task a process for pings/cleanups
        self._loop.create_task(self._handler())

    async def stop(self) -> None:
        """Stop connection to SniTun server."""
        if not self._multiplexer:
            raise RuntimeError("No SniTun connection available")
        await self._multiplexer.shutdown()

    async def _handler(self) -> None:
        """Wait until connection is closed."""
        try:
            while self._multiplexer.is_connected:
                try:
                    async with async_timeout.timeout(50):
                        await self._multiplexer.wait()
                except asyncio.TimeoutError:
                    self._multiplexer.ping()
        finally:
            self._multiplexer = None



"""Utils for server handling."""
from datetime import datetime, timedelta
from typing import List
import json

from cryptography.fernet import Fernet, MultiFernet


def generate_client_token(tokens: List[str], valid_delta: timedelta,
                          hostname: str, aes_key: bytes, aes_iv: bytes):
    """Generate a token for client."""
    fernet = MultiFernet([Fernet(key) for key in tokens])
    valid = datetime.utcnow() + valid_delta

    return fernet.encrypt(
        json.dumps({
            "valid": valid.timestamp(),
            "hostname": hostname,
            "aes_key": aes_key.hex(),
            "aes_iv": aes_iv.hex(),
        }).encode())



"""TLS ClientHello parser."""
import logging

from ..exceptions import ParseSNIError

_LOGGER = logging.getLogger(__name__)

TLS_HEADER_LEN = 5
TLS_HANDSHAKE_CONTENT_TYPE = 0x16
TLS_HANDSHAKE_TYPE_CLIENT_HELLO = 0x01


def parse_tls_sni(data: bytes) -> str:
    """Parse TLS SNI extention."""

    if len(data) < TLS_HEADER_LEN:
        _LOGGER.debug("Invalid TLS header")
        raise ParseSNIError()

    # If TLS handshake
    if data[0] != TLS_HANDSHAKE_CONTENT_TYPE:
        _LOGGER.debug("Not TLS handshake received")
        raise ParseSNIError()

    # Check compatible ClientHello
    if int(data[1]) < 3:
        _LOGGER.debug("Received ClientHello without SNI support")
        raise ParseSNIError()

    # Calculate TLS record size
    tls_size = (data[3] << 8) + data[4] + TLS_HEADER_LEN
    if len(data) < tls_size:
        _LOGGER.debug("Can't calculate the TLS record size")
        raise ParseSNIError()

    # Check if handshake is a ClientHello
    pos = TLS_HEADER_LEN
    if data[pos] != TLS_HANDSHAKE_TYPE_CLIENT_HELLO:
        _LOGGER.debug("Invalid ClientHello type")
        raise ParseSNIError()

    # Seek fixed length header part
    pos += 38

    # Seek SessionID
    try:
        pos += 1 + data[pos]
    except IndexError:
        _LOGGER.debug("Invalid SessionID")
        raise ParseSNIError() from None

    # Seek Cipher Suites
    try:
        pos += 2 + (data[pos] << 8) + data[pos + 1]
    except IndexError:
        _LOGGER.debug("Invalid CipherSuites")
        raise ParseSNIError() from None

    # Seek Compression Methods
    try:
        pos += 1 + data[pos]
    except IndexError:
        _LOGGER.debug("Invalid CompressionMethods")
        raise ParseSNIError() from None

    # Check data buffer + extension size
    if pos + 2 > len(data):
        _LOGGER.debug("Mismatch Extension TLS header")
        raise ParseSNIError()

    # Process extension
    return _parse_extension(data, pos)


def _parse_extension(data: bytes, pos: int) -> str:
    """Parse TLS ClientHello Extension."""

    # Seek Extension start
    try:
        tls_extension_size = (data[pos] << 8) + data[pos + 1]
        pos += 2
    except IndexError:
        raise ParseSNIError() from None

    # Check data buffer + extension size
    if pos + tls_extension_size > len(data):
        _LOGGER.debug("Mismatch Extension TLS header")
        raise ParseSNIError()

    # Loop over extension until we have our SNI
    while pos + 4 <= len(data):

        # SNI?
        if data[pos] == 0x00 and data[pos + 1] == 0x00:
            return _parse_host_name(data, pos + 4)

        pos += 4 + (data[pos + 2] << 8) + data[pos + 3]

    _LOGGER.debug("Can't find any ServerName Extension")
    raise ParseSNIError()


def _parse_host_name(data: bytes, pos: int) -> str:
    """Parse TLS ServerName Extension."""

    # Seek list size
    pos += 2

    while pos + 3 < len(data):
        size = (data[pos + 1] << 8) + data[pos + 2]

        # Unknown server name type
        if data[pos] != 0x00:
            _LOGGER.debug("Unknown ServerName type")
            pos += 3 + size
            continue

        try:
            return bytes(data[pos + 3:pos + 3 + size]).decode()
        except IndexError:
            _LOGGER.debug("Wrong host length")
            raise ParseSNIError() from None

    _LOGGER.debug("Not found any valid ServerName")
    raise ParseSNIError()
